/****** joysticks.acepansion/--background-- *********************************
*
* DESCRIPTION
*  This plugin provides joysticks emulation to ACE CPC Emulator.
*
* HISTORY
*
*****************************************************************************
*
*/

#include <clib/alib_protos.h>

#include <proto/exec.h>
#include <proto/utility.h>

#define USE_INLINE_API
#include <acepansion/lib_header.h>
#include <acepansion/plugin.h>

#include "acepansion.h"
#include "interface.h"
#define CATCOMP_NUMBERS
#include "generated/locale_strings.h" /* catalog string ids */



struct Library *UtilityBase;



struct PluginDataJoy
{
    ULONG idx;
    ULONG analogMode;
    STRPTR preferredName;

    USHORT buttons;
    BYTE ns;
    BYTE ew;
    BYTE lx;
    BYTE ly;
    BYTE rx;
    BYTE ry;
};

struct PluginData
{
    // Plugin generic part (do NEVER put anything before this!)
    struct ACEpansionPlugin common;

    // Joysticks specifics
    struct SignalSemaphore *sema;
    STRPTR *list;
    ULONG combinedMode;
    struct PluginDataJoy joy[2];
};


#define TOOLTYPE_JOY0     "JOYSTICK0"
#define TOOLTYPE_JOY1     "JOYSTICK1"
#define TOOLTYPE_ANALOG0  "ANALOG0"
#define TOOLTYPE_ANALOG1  "ANALOG1"
#define TOOLTYPE_COMBINED "COMBINED"

/*
** Private function which is called to initialize commons
** just after the library was loaded into memory and prior
** to any other API call
**
** This is the good place to open our required libraries
*/
BOOL InitResources(VOID)
{
    UtilityBase = OpenLibrary("utility.library", 0L);

    return UtilityBase && GUI_InitResources();
}

/*
** Private function which is called to free commons
** just before the library is expurged from memory
**
** This is the good place to close our required libraries
*/
VOID FreeResources(VOID)
{
    GUI_FreeResources();

    CloseLibrary(UtilityBase);
}



/****** joysticks.library/CreatePlugin **************************************
*
* NAME
*   CreatePlugin -- (V1) -- GUI
*
* SYNOPSIS
*   struct ACEpansionPlugin * CreatePlugin(CONST_STRPTR *toolTypes, struct SignalSemaphore *sema)
*
* FUNCTION
*   Create the plugin data structure and configure how it will be handled by
*   ACE. This function is usually called at ACE startup.
*
* INPUTS
*   *toolType
*     A pointer to a NULL terminated array of CONST_STRPTR containing the
*     tooltypes provided with the plugin library.
*   *sema
*     Pointer to a semaphore to use to protect access to custom data from
*     ACEpansionPlugin when accessed from subtasks or within your PrefsWindow.
*
* RESULT
*     A pointer to an allocated and initialized plugin data structure.
*
* SEE ALSO
*   DeletePlugin
*   ActivatePlugin
*   GetPrefsPlugin
*
*****************************************************************************
*
*/

ACEPANSION_API
struct ACEpansionPlugin * CreatePlugin(CONST_STRPTR *toolTypes, struct SignalSemaphore *sema)
{
    // We are allocating an extended plugin structure to add our custom stuff
    struct PluginData *myPlugin = (struct PluginData*)AllocVec(sizeof(struct PluginData), MEMF_PUBLIC|MEMF_CLEAR);

    if(myPlugin)
    {
        // Get tooltypes to configure ourself
        if(toolTypes) while(*toolTypes)
        {
            CONST_STRPTR ptr = *toolTypes;
            CONST_STRPTR name = ptr;
            ULONG len = 0;

            if(*ptr != '(')
            {
                while(*ptr != '=' && *ptr != '\0')
                {
                    ptr++;
                    len++;
                }

                if(*ptr == '=')
                {
                    // Parameters
                    if(Strnicmp(name,TOOLTYPE_JOY0,len) == 0)
                    {
                        CONST_STRPTR value = ptr+1;

                        if(*value != '\0')
                        {
                            myPlugin->joy[0].preferredName = StrDup(value);
                        }
                    }
                    else if(Strnicmp(name,TOOLTYPE_JOY1,len) == 0)
                    {
                        CONST_STRPTR value = ptr+1;

                        if(*value != '\0')
                        {
                            myPlugin->joy[1].preferredName = StrDup(value);
                        }
                    }
                }
                else
                {
                    // Switches
                    if(Strnicmp(name,TOOLTYPE_ANALOG0,len) == 0)
                    {
                        myPlugin->joy[0].analogMode = TRUE;
                    }
                    else if(Strnicmp(name,TOOLTYPE_ANALOG1,len) == 0)
                    {
                        myPlugin->joy[1].analogMode = TRUE;
                    }
                    else if(Strnicmp(name,TOOLTYPE_COMBINED,len) == 0)
                    {
                        myPlugin->combinedMode = TRUE;
                    }
                }
            }
            toolTypes++;
        }

        // Initialize plugin specifics
        myPlugin->sema = sema;

        // Initialize plugin generic part
        myPlugin->common.ap_APIVersion     = API_VERSION;
        myPlugin->common.ap_Flags          = ACE_FLAGSF_ACTIVE_JOYSTICK     // We need the joysticks API (but we are not exclusive)
                                           | ACE_FLAGSF_ACTIVE_ANALOG_INPUT // We handle Amstrad Plus analog joysticks
                                           | ACE_FLAGSF_SAVE_PREFS          // We have preferences to save
                                           | ACE_FLAGSF_CATCH_HOST_GAMEPAD; // We use gamepads plugged on the host computer
        myPlugin->common.ap_Title          = GetString(MSG_TITLE);
        myPlugin->common.ap_HelpFileName   = LIBNAME".guide";
        myPlugin->common.ap_ToggleMenuName = GetString(MSG_MENU_TOGGLE);
        myPlugin->common.ap_PrefsMenuName  = GetString(MSG_MENU_PREFS);
        myPlugin->common.ap_PrefsWindow    = GUI_Create((struct ACEpansionPlugin *)myPlugin);

        if(myPlugin->common.ap_PrefsWindow)
        {
            GUI_SetJoystickCombinedMode(myPlugin->common.ap_PrefsWindow, myPlugin->combinedMode);
            GUI_SetJoystickAnalogMode(myPlugin->common.ap_PrefsWindow, 0, myPlugin->joy[0].analogMode);
            GUI_SetJoystickAnalogMode(myPlugin->common.ap_PrefsWindow, 1, myPlugin->joy[1].analogMode);
        }
        else
        {
            DeletePlugin((struct ACEpansionPlugin *)myPlugin);
            myPlugin = NULL;
        }
    }

    return (struct ACEpansionPlugin *)myPlugin;
}



/****** joysticks.library/DeletePlugin **************************************
*
* NAME
*   DeletePlugin -- (V1) -- GUI
*
* SYNOPSIS
*   VOID DeletePlugin(struct ACEpansionPlugin *plugin)
*
* FUNCTION
*   Delete a plugin data structure. This function is usually called when ACE
*   is exiting.
*
* INPUTS
*   *plugin
*     Pointer to the plugin data structure to delete.
*
* RESULT
*
* SEE ALSO
*   CreatePlugin
*   ActivatePlugin
*
*****************************************************************************
*
*/

ACEPANSION_API
VOID DeletePlugin(struct ACEpansionPlugin *plugin)
{
    struct PluginData *myPlugin = (struct PluginData *)plugin;

    FreeVec(myPlugin->joy[0].preferredName);
    FreeVec(myPlugin->joy[1].preferredName);
    GUI_Delete(myPlugin->common.ap_PrefsWindow);
    FreeVec(myPlugin);
}



/****** joysticks.library/ActivatePlugin ************************************
*
* NAME
*   ActivatePlugin -- (V1) -- GUI
*
* SYNOPSIS
*   BOOL ActivatePlugin(struct ACEpansionPlugin *plugin, BOOL status)
*
* FUNCTION
*   This function is called everytime ACE wants to activate or disactivate your plugin.
*   While you are disactivated, no other function (except Delete) will the called by ACE.
*   Please note that a plugin is always disactivated by ACE prior to DeletePlugin().
*
* INPUTS
*   activate
*     New activation status requested by ACE.
*
* RESULT
*   Actual activation status (you may fail at activation for some reason, but you
*   should never fail at disactivation request).
*
* SEE ALSO
*   CreatePlugin
*   DeletePlugin
*
*****************************************************************************
*
*/

ACEPANSION_API
BOOL ActivatePlugin(UNUSED struct ACEpansionPlugin *plugin, BOOL activate)
{
    return activate;
}



/****** joysticks.acepansion/GetPrefsPlugin *********************************
*
* NAME
*   GetPrefsPlugin -- (V5) -- GUI
*
* SYNOPSIS
*   STRPTR * GetPrefsPlugin(struct ACEpansionPlugin *plugin)
*
* FUNCTION
*   Function that is used by ACE to know about the current preferences of the
*   plugins in order to save them.
*
* INPUTS
*   *plugin
*     Pointer to the plugin data structure to use.
*   pool
*     Memory pool to use to allocate the tooltypes.
*
* RESULT
*   Pointer to a NULL terminated string array containing the tooltypes or NULL
*   when no tooltypes.
*
* SEE ALSO
*   CreatePlugin
*
*****************************************************************************
*
*/

ACEPANSION_API
STRPTR * GetPrefsPlugin(struct ACEpansionPlugin *plugin, APTR pool)
{
    struct PluginData *myPlugin = (struct PluginData *)plugin;
    STRPTR *tooltypes = (STRPTR*)AllocPooled(pool, 6*sizeof(STRPTR));

    if(tooltypes)
    {
        LONG i = 0;

        // When prefs are requested, we have to freeze
        // the current selection as the preferred one
        FreeVec(myPlugin->joy[0].preferredName);
        if(myPlugin->list && myPlugin->joy[0].idx)
            myPlugin->joy[0].preferredName = StrDup(myPlugin->list[myPlugin->joy[0].idx]);
        else
            myPlugin->joy[0].preferredName = NULL;

        FreeVec(myPlugin->joy[1].preferredName);
        if(myPlugin->joy[1].idx)
            myPlugin->joy[1].preferredName = StrDup(myPlugin->list[myPlugin->joy[1].idx]);
        else
            myPlugin->joy[1].preferredName = NULL;

        // Then we provide it
        if(myPlugin->joy[0].preferredName) tooltypes[i++] = MakeToolType(pool, TOOLTYPE_JOY0    , myPlugin->joy[0].preferredName);
        if(myPlugin->joy[0].analogMode)    tooltypes[i++] = MakeToolType(pool, TOOLTYPE_ANALOG0 , NULL                          );
        if(myPlugin->joy[1].preferredName) tooltypes[i++] = MakeToolType(pool, TOOLTYPE_JOY1    , myPlugin->joy[1].preferredName);
        if(myPlugin->joy[1].analogMode)    tooltypes[i++] = MakeToolType(pool, TOOLTYPE_ANALOG1 , NULL                          );
        if(myPlugin->combinedMode)         tooltypes[i++] = MakeToolType(pool, TOOLTYPE_COMBINED, NULL                          );

        tooltypes[i++] = NULL;
    }

    return tooltypes;
}



/****** joysticks.library/Reset *********************************************
*
* NAME
*   Reset -- (V1) -- EMU
*
* SYNOPSIS
*   VOID Reset(struct ACEpansionPlugin *plugin)
*
* FUNCTION
*   Function that is called everytime ACE is issuing a bus reset.
*
* INPUTS
*   *plugin
*     Pointer to the plugin data structure to use.
*
* RESULT
*
* SEE ALSO
*   Emulate
*
*****************************************************************************
*
*/

ACEPANSION_API
VOID Reset(UNUSED struct ACEpansionPlugin *plugin)
{
}



/****** joysticks.library/Emulate *******************************************
*
* NAME
*   Emulate -- (V1) -- EMU
*
* SYNOPSIS
*   VOID Emulate(struct ACEpansionPlugin *plugin, ULONG *signals)
*
* FUNCTION
*   Function that is called at each microsecond of the emulation step inside ACE.
*   Note: Implement this function only if is it really required for your plugin because it
*   is obviously CPU intensive.
*
* INPUTS
*   *plugin
*     Pointer to the plugin data structure to use.
*   *signals
*     Pointer to the I/O signals from the expansion port you can use or alter
*     depending on what you are emulating.
*     The signals bits can be decoded using the following flags:
*       ACE_SIGNALF_INT -- (V1)
*         Maskable interrupt; corresponding to /INT signal from CPC expansion
*         port.
*       ACE_SIGNALF_NMI -- (V3)
*         Non-maskable interrupt; corresponding to /NMI signal from CPC
*         expansion port.
*       ACE_SIGNALF_LPEN -- (V4)
*         Light pen trigger; corresponding to /LPEN signal from CPC expansion
*         port.
*       ACE_SIGNALF_BUS_RESET -- (V6)
*         Bus reset trigger; correspond to /BUSRESET signal from CPC expansion
*         port.
*       ACE_SIGNALF_NOT_READY -- (V6)
*         Not ready trigger; correspond to READY signal from CPC expansion port.
*       ACE_SIGNALF_PLUS_BUS -- (V6)
*         Read-only flag indicating that the expansion port is an Amstrad Plus
*         one (i.e. providing a 4.4444 MHz clock instead of 4 MHz).
*
* RESULT
*
* SEE ALSO
*   Reset
*
*****************************************************************************
*
*/

ACEPANSION_API
VOID Emulate(UNUSED struct ACEpansionPlugin *plugin, UNUSED ULONG *signals)
{
}



/****** joysticks.library/WriteIO *******************************************
*
* NAME
*   WriteIO -- (V1) -- EMU
*
* SYNOPSIS
*   VOID WriteIO(struct ACEpansionPlugin *plugin, USHORT port, UBYTE value)
*
* FUNCTION
*   Function that is called everytime Z80 is performing a I/O port write operation.
*   Note: Implement this function only if is it really required for your plugin because it
*   is obviously CPU intensive.
*
* INPUTS
*   port
*     Z80 port on with the I/O write operation was issued.
*   value
*     Value that was issued on the port.
*
* RESULT
*
* SEE ALSO
*   ReadIO
*
*****************************************************************************
*
*/

ACEPANSION_API
VOID WriteIO(UNUSED struct ACEpansionPlugin *plugin, UNUSED USHORT port, UNUSED UBYTE value)
{
}



/****** joysticks.library/ReadIO ********************************************
*
* NAME
*   ReadIO -- (V1) -- EMU
*
* SYNOPSIS
*   VOID ReadIO(struct ACEpansionPlugin *plugin, USHORT port, UBYTE *value)
*
* FUNCTION
*   Function that is called everytime Z80 is performing a I/O port read operation.
*   Note: Implement this function only if is it really required for your plugin because it
*   is obviously CPU intensive.
*
* INPUTS
*   port
*     Z80 port on with the I/O read operation was issued.
*   *value
*     Value to be returned.
*
* RESULT
*
* SEE ALSO
*   WriteIO
*
*****************************************************************************
*
*/

ACEPANSION_API
VOID ReadIO(UNUSED struct ACEpansionPlugin *plugin, UNUSED USHORT port, UNUSED UBYTE *value)
{
}



/****** joystick.acepansion/WriteMem ****************************************

* NAME
*  WriteMem -- (V6) -- EMU
*
* SYNOPSIS
*  BOOL WriteMem(struct ACEpansionPlugin *plugin, USHORT address, UBYTE value)
*
* FUNCTION
*  Function that is called everytime Z80 is performing a memory write
*  operation.
*
* INPUTS
*  *plugin
*    Pointer to the plugin data structure to use.
*  address
*    Z80 address on which the memory write operation was issued.
*  value
*    Value that has to be written.
*
* RESULT
*  TRUE if the plugins actually cougth the memory write operation so that
*  internal memory will be left unchanged (/RAMDIS is set to 0).
*
* NOTES
*  Implement this function only if is it really required for your plugin
*  because it is obviously CPU intensive.
*
* SEE ALSO
*  ReadMem()
*
*****************************************************************************
*
*/

ACEPANSION_API
BOOL WriteMem(UNUSED struct ACEpansionPlugin *plugin, UNUSED USHORT address, UNUSED UBYTE value)
{
    return FALSE;
}



/****** joysticks.acepansion/ReadMem ****************************************
*
* NAME
*  ReadMem -- (V6) -- EMU
*
* SYNOPSIS
*  VOID ReadMem(struct ACEpansionPlugin *plugin, USHORT address, UBYTE *value, BOOL opcodeFetch) -- (V7)
*  BOOL ReadMem(struct ACEpansionPlugin *plugin, USHORT address, UBYTE *value, BOOL opcode) ** DEPRECATED in V7 **

* FUNCTION
*  Function that is called everytime Z80 is performing a memory read
*  operation.
*
* INPUTS
*  *plugin
*    Pointer to the plugin data structure to use.
*  address
*    Z80 address on which memory read operation was issued.
*  *value
*    Value to be returned.
*    Starting with V7, it contains by default the value that might be read by
*    the CPC if the plugin won't catch the memory read, so that a plugin can
*    only snif the bus.
*  opcodeFetch
*    Boolean which is TRUE when the memory read operation is related to
*    Z80 opcode fetching (/M1 signal). If FALSE, then this is a regular
*    read operation during opcode execution.
*
* RESULT
*  ** DEPRECATED in V7 **
*  TRUE if the plugins actually cougth the memory read operation so that
*  internal memory won't be read (both /ROMDIS and /RAMDIS are set to 0).
*
* NOTES
*  Implement this function only if is it really required for your plugin
*  because it is obviously CPU intensive.
*
* SEE ALSO
*  WriteMem()
*
*****************************************************************************
*
*/

ACEPANSION_API
VOID ReadMem(UNUSED struct ACEpansionPlugin *plugin, UNUSED USHORT address, UNUSED UBYTE *value, UNUSED BOOL opcodeFetch)
{
}



/****** joysticks.acepansion/GetAudio ***************************************
*
* NAME
*   GetAudio -- (V1) -- EMU
*
* SYNOPSIS
*   VOID GetAudio(struct ACEpansionPlugin *plugin, SHORT *leftSample, SHORT *rightSample)
*
* FUNCTION
*   Function to be used when your plugin needs to mix an audio output with the one
*   from ACE.
*   It is called at each PSG emulation step, which means at 125KHz.
*
* INPUTS
*   *leftSample
*     Pointer to store the left audio sample to be mixed with ACE audio.
*   *rightSample
*     Pointer to store the right audio sample to be mixed with ACE audio.
*
* RESULT
*
* SEE ALSO
*
*****************************************************************************
*
*/

ACEPANSION_API
VOID GetAudio(UNUSED struct ACEpansionPlugin *plugin, UNUSED SHORT *leftSample, UNUSED SHORT *rightSample)
{
}



/****** joysticks.library/Printer *******************************************
*
* NAME
*   Printer -- (V1) -- EMU
*
* SYNOPSIS
*   VOID WritePrinter(struct ACEpansionPlugin *plugin, UBYTE data, BOOL strobe, BOOL *busy)
*
* FUNCTION
*   Function that is called everytime the CPC is accessing the printer port.
*
* INPUTS
*   data
*     8 bits written value (note that CPC is limited to 7 bits values, only Amstrad Plus
*     can actually provide the 8th bit).
*   strobe
*     Status of the /STROBE signal from the printer port.
*   *busy
*     Busy signal value that can be altered.
*
* RESULT
*
* SEE ALSO
*
*****************************************************************************
*
*/

ACEPANSION_API
VOID Printer(UNUSED struct ACEpansionPlugin *plugin, UNUSED UBYTE data, UNUSED BOOL strobe, UNUSED BOOL *busy)
{
}



/****** joysticks.library/Joystick ******************************************
*
* NAME
*   Joystick -- (V2) -- EMU
*
* SYNOPSIS
*   VOID Joystick(struct ACEpansionPlugin *plugin, BOOL com1, BOOL com2, UBYTE *ioData)
*
* FUNCTION
*   Function which is called everytime the CPC is accessing the joystick port.
*   Please note that the joystick port in R/W but you cannot know if the CPC
*   is actually reading or writing to it.
*
* INPUTS
*   com1
*     If com1 is active then the CPC is requesting access to joystick 0.
*   com2
*     If com2 is active then the CPC is requesting access to joystick 1.
*   *ioData
*     If the CPC is reading the joystick port (regular case) ioData will be set
*     to 0xff and you are supposed to set to 0 the bits corresponding to the
*     requested joystick (you should use the ACE_IODATAF_JOYSTICK_XXX defines).
*     Please note the additional ACE_IODATAF_JOYSTICK_PAUSE which is not
*     existing on CPC but that you could use to map the play/pause button of
*     a joystick used in ACE on the 'P' key of the keyboard (or the GX4000
*     pause button).
*     If the CPC is writing the joystick port the ioData will contain the
*     written value to the port and modifying it will have not effect (on a
*     real CPC it could simply destroy the PSG chipset!).
*
* RESULT
    To emulate the analog joystick port of the Amstrad Plus and the GX-4000,
    this function should be used together with AnalogInput().
*
* SEE ALSO
*
*****************************************************************************
*
*/

ACEPANSION_API
VOID Joystick(struct ACEpansionPlugin *plugin, BOOL com1, BOOL com2, UBYTE *ioData)
{
    struct PluginData *myPlugin = (struct PluginData *)plugin;
    struct PluginDataJoy *joy = NULL;

    if(!myPlugin->combinedMode)
    {
        // This is the legacy mode where one host's gamepad controls one CPC joysticks
        if(com1)
        {
            joy = &myPlugin->joy[0];
        }
        else if(com2)
        {
            joy = &myPlugin->joy[1];
        }

        if(joy)
        {
            // Buttons mapping
            if(joy->buttons & ACE_HOSTGAMEPADF_BUTTON_LEFT)
                *ioData &= ~ACE_IODATAF_JOYSTICK_FIRE1;
            if(joy->buttons & ACE_HOSTGAMEPADF_BUTTON_BOTTOM)
                *ioData &= ~ACE_IODATAF_JOYSTICK_FIRE2;
            if(joy->buttons & (ACE_HOSTGAMEPADF_BUTTON_TOP|ACE_HOSTGAMEPADF_BUTTON_LEFT_STICK|ACE_HOSTGAMEPADF_BUTTON_RIGHT_STICK))
                *ioData &= ~ACE_IODATAF_JOYSTICK_FIRE3;
            // Add a mapping to up direction from a button to jump easily in some games
            if(joy->buttons & ACE_HOSTGAMEPADF_BUTTON_RIGHT)
                *ioData &= ~ACE_IODATAF_JOYSTICK_UP;
            // Pause from gamepad
            if(joy->buttons & (ACE_HOSTGAMEPADF_BUTTON_MIDDLE_LEFT|ACE_HOSTGAMEPADF_BUTTON_MIDDLE_RIGHT))
                *ioData &= ~ACE_IODATAF_JOYSTICK_PAUSE;
            // East-West and L1/R1 are always mapped to left/right, LX and RX only when no analog mode
            if(joy->buttons & ACE_HOSTGAMEPADF_BUTTON_UPPER_REAR_RIGHT || joy->ew > +32 || (!joy->analogMode && (joy->lx > +32 || joy->rx > +32)))
                *ioData &= ~ACE_IODATAF_JOYSTICK_RIGHT;
            if(joy->buttons & ACE_HOSTGAMEPADF_BUTTON_UPPER_REAR_LEFT || joy->ew < -32 || (!joy->analogMode && (joy->lx < -32 || joy->rx < -32)))
                *ioData &= ~ACE_IODATAF_JOYSTICK_LEFT;
            // North-South and L2/R2 are always mapped to up/down, LY and RY only when no analog mode
            if(joy->buttons & ACE_HOSTGAMEPADF_BUTTON_LOWER_REAR_RIGHT || joy->ns > +32 || (!joy->analogMode && (joy->ly > +32 || joy->ry > +32)))
                *ioData &= ~ACE_IODATAF_JOYSTICK_DOWN;
            if(joy->buttons & ACE_HOSTGAMEPADF_BUTTON_LOWER_REAR_LEFT || joy->ns <-32 || (!joy->analogMode && (joy->ly < -32 || joy->ry < -32)))
                *ioData &= ~ACE_IODATAF_JOYSTICK_UP;
        }
    }
    else
    {
        // This is the enhanced mode where one host's gamepad controls both CPC joysticks
        joy = &myPlugin->joy[0];

        if(com1)
        {
            // Buttons mapping
            if(joy->buttons & ACE_HOSTGAMEPADF_BUTTON_LEFT)
                *ioData &= ~ACE_IODATAF_JOYSTICK_FIRE1;
            if(joy->buttons & ACE_HOSTGAMEPADF_BUTTON_BOTTOM)
                *ioData &= ~ACE_IODATAF_JOYSTICK_FIRE2;
            if(joy->buttons & ACE_HOSTGAMEPADF_BUTTON_LEFT_STICK)
                *ioData &= ~ACE_IODATAF_JOYSTICK_FIRE3;
            // Pause from gamepad
            if(joy->buttons & (ACE_HOSTGAMEPADF_BUTTON_MIDDLE_LEFT|ACE_HOSTGAMEPADF_BUTTON_MIDDLE_RIGHT))
                *ioData &= ~ACE_IODATAF_JOYSTICK_PAUSE;
            // East-West are always mapped to joy0 left/right, X only when no analog mode
            if(joy->ew > +32 || (!joy->analogMode && joy->lx > +32))
                *ioData &= ~ACE_IODATAF_JOYSTICK_RIGHT;
            if(joy->ew < -32 || (!joy->analogMode && joy->lx < -32))
                *ioData &= ~ACE_IODATAF_JOYSTICK_LEFT;
            // North-South are always mapped to joy 0 up/down, Y only when no analog mode
            if(joy->ns > +32 || (!joy->analogMode && joy->ly > +32))
                *ioData &= ~ACE_IODATAF_JOYSTICK_DOWN;
            if(joy->ns <-32 || (!joy->analogMode && joy->ly < -32))
                *ioData &= ~ACE_IODATAF_JOYSTICK_UP;
        }
        else if(com2)
        {
            // Buttons mapping
            if(joy->buttons & ACE_HOSTGAMEPADF_BUTTON_TOP)
                *ioData &= ~ACE_IODATAF_JOYSTICK_FIRE1;
            if(joy->buttons & ACE_HOSTGAMEPADF_BUTTON_RIGHT)
                *ioData &= ~ACE_IODATAF_JOYSTICK_FIRE2;
            if(joy->buttons & ACE_HOSTGAMEPADF_BUTTON_RIGHT_STICK)
                *ioData &= ~ACE_IODATAF_JOYSTICK_FIRE3;
            // Pause from gamepad
            if(joy->buttons & (ACE_HOSTGAMEPADF_BUTTON_MIDDLE_LEFT|ACE_HOSTGAMEPADF_BUTTON_MIDDLE_RIGHT))
                *ioData &= ~ACE_IODATAF_JOYSTICK_PAUSE;
            // L1 and R1 are always mapped to joy1 left/right, Z only when no analog mode
            if(joy->buttons & ACE_HOSTGAMEPADF_BUTTON_UPPER_REAR_RIGHT || (!joy->analogMode && joy->rx > +32))
                *ioData &= ~ACE_IODATAF_JOYSTICK_RIGHT;
            if(joy->buttons & ACE_HOSTGAMEPADF_BUTTON_UPPER_REAR_LEFT || (!joy->analogMode && joy->rx < -32))
                *ioData &= ~ACE_IODATAF_JOYSTICK_LEFT;
            // L2 and R2 are always mapped to joy1 up/down, Rot only when no analog mode
            if(joy->buttons & ACE_HOSTGAMEPADF_BUTTON_LOWER_REAR_RIGHT || (!joy->analogMode && joy->ry > +32))
                *ioData &= ~ACE_IODATAF_JOYSTICK_DOWN;
            if(joy->buttons & ACE_HOSTGAMEPADF_BUTTON_LOWER_REAR_LEFT || (!joy->analogMode && joy->ry < -32))
                *ioData &= ~ACE_IODATAF_JOYSTICK_UP;
        }
    }
}



/****** joysticks.acepansion/AnalogInput ************************************
*
* NAME
*   AnalogInput -- (V7) -- EMU
*
* SYNOPSIS
*   VOID AnalogInput(struct ACEpansionPlugin *plugin, UBYTE channel, UBYTE *value)
*
* FUNCTION
*   Function which is called everytime the CPC is accessing the analog joystick
*   port of the Amstrad Plus or the GX-4000.
*
* INPUTS
*   *plugin
*     Pointer to the plugin data structure to use.
*   channel
*     Analog channel number (from 0 to 3).
*   *value
*     Pointer where to store the 6-bit wide value of the data on the channel.
*
* RESULT
*
* NOTES
*   To emulate the analog joystick port of the Amstrad Plus and the GX-4000,
*   this function should be used together with Joystick() which is providing
*   fire buttons (they are shared between both analog and digital joystick
*   ports).
*
* SEE ALSO
*   Joystick()
*
*****************************************************************************
*
*/

ACEPANSION_API
VOID AnalogInput(struct ACEpansionPlugin *plugin, UBYTE channel, UBYTE *value)
{
    struct PluginData *myPlugin = (struct PluginData *)plugin;

    switch(channel)
    {
    case 0:
        if(myPlugin->joy[0].analogMode)
            *value = (myPlugin->joy[0].lx + 128) / 4;
        break;
    case 1:
        if(myPlugin->joy[0].analogMode)
            *value = (myPlugin->joy[0].ly + 128) / 4;
        break;
    case 2:
        if(myPlugin->joy[0].analogMode && myPlugin->combinedMode)
            *value = (myPlugin->joy[0].rx + 128) / 4;
        else if(myPlugin->joy[1].analogMode && !myPlugin->combinedMode)
            *value = (myPlugin->joy[1].lx + 128) / 4;
        break;
    case 3:
        if(myPlugin->joy[0].analogMode && myPlugin->combinedMode)
            *value = (myPlugin->joy[0].ry + 128) / 4;
        else if(myPlugin->joy[1].analogMode && !myPlugin->combinedMode)
            *value = (myPlugin->joy[1].ly + 128) / 4;
        break;
    }
}



/****** joysticks.library/AcknowledgeInterrupt ******************************
*
* NAME
*   AcknowledgeInterrupt -- (V1) -- EMU
*
* SYNOPSIS
*   VOID AcknowledgeInterrupt(struct ACEpansionPlugin *plugin, UBYTE *ivr)
*
* FUNCTION
*   Function which is called everytime the Z80 is entering an interrupt.
*
* INPUTS
*   *ivr
*     IVR could be set to the desired value (Interrupt Vector Register).
*
* RESULT
*
* SEE ALSO
*   ReturnInterrupt
*
*****************************************************************************
*
*/

ACEPANSION_API
VOID AcknowledgeInterrupt(UNUSED struct ACEpansionPlugin *plugin, UNUSED UBYTE *ivr)
{
}



/****** joysticks.acepansion/ReturnInterrupt ********************************
*
* NAME
*   ReturnInterrupt -- (V3) -- EMU
*
* SYNOPSIS
*   VOID ReturnInterrupt(struct ACEpansionPlugin *plugin)
*
* FUNCTION
*   Function which is called everytime the Z80 is returning from an interrupt
*   using RETI instruction.
*
* INPUTS
*
* RESULT
*
* SEE ALSO
*   AcknowledgeInterrupt
*
*****************************************************************************
*
*/

ACEPANSION_API
VOID ReturnInterrupt(UNUSED struct ACEpansionPlugin *plugin)
{
}



/****** joysticks.acepansion/Cursor *****************************************
*
* NAME
*   Cursor -- (V1) -- EMU
*
* SYNOPSIS
*   VOID Cursor(struct ACEpansionPlugin *plugin)
*
* FUNCTION
*   Function which is called everytime the cursor signal from the CRTC is
*   active.
*
* INPUTS
*
* RESULT
*
* SEE ALSO
*
*****************************************************************************
*
*/

ACEPANSION_API
VOID Cursor(UNUSED struct ACEpansionPlugin *plugin)
{
}



/****** joysticks.acepansion/HostGamepadList ********************************
*
* NAME
*   HostGamepadList -- (V7) -- GUI

* SYNOPSIS
*   VOID HostGamepadList(struct ACEpansionPlugin *plugin, STRPTR *list)
*
* FUNCTION
*   This function is called everytime a gamepad in plugged or unplugged from
*   the host so that the plugin can know about the available devices. Index
*   from HostGamepadEvent() will refer to this list.
*
* INPUTS
*   *plugin
*     Pointer to the plugin data structure to use.
*   list
*     NULL terminated array of strings containing the human readable names of
*     the plugged gamepads, the first one being the localized string for "No
*     joystick". Please note that the contents of this list remains valid
*     until the function is invoked again, so that you can use it safely in
*     your code without duplication.
*
* RESULT
*
* NOTES
*
* SEE ALSO
*   HostGamepadList()
*
*****************************************************************************
*
*/

ACEPANSION_API
VOID HostGamepadList(struct ACEpansionPlugin *plugin, STRPTR *list)
{
    struct PluginData *myPlugin = (struct PluginData *)plugin;
    LONG idx;
    LONG joy0 = 0;
    LONG joy1 = 0;

    for(idx=1; list[idx] != NULL; idx++)
    {
        // Check if we are matching preferences
        if((joy0 == 0 && idx == 1) // Default to 1st found joytick
        || (myPlugin->joy[0].preferredName && Stricmp(myPlugin->joy[0].preferredName, list[idx]) == 0))
        {
            joy0 = idx;
        }
        if((joy1 == 0 && idx == 2) // Default to 2nd found joytick
        || (myPlugin->joy[1].preferredName && Stricmp(myPlugin->joy[1].preferredName, list[idx]) == 0))
        {
            joy1 = idx;
        }
    }

    myPlugin->list = list;
    myPlugin->joy[0].idx = joy0;
    myPlugin->joy[1].idx = joy1;

    GUI_SetJoystickList(myPlugin->common.ap_PrefsWindow, list, joy0, joy1);
}



/****** joysticks.acepansion/HostGamepadEvent *******************************
*
* NAME
*   HostGamepadEvent -- (V7) -- GUI
*
* SYNOPSIS
*   VOID HostGamepadEvent(struct ACEpansionPlugin *plugin, UBYTE index, USHORT buttons, BYTE ns, BYTE ew, BYTE lx, BYTE ly, BYTE rx, BYTE ry)
*
* FUNCTION
*   This function is called everytime the state of a host's gamepad state is
*   changing. It let you handle both analogic and digital devices.
*
* INPUTS
*   *plugin
*     Pointer to the plugin data structure to use.
*   index
*     Index of the gamepad in gamepad list.
*   buttons
*     State of the gamepad buttons.
*     Buttons bits can be decoded using the ACE_HOSTGAMEPADF_BUTTON_XXX flags.
*   ns
*     Horizontal position of directional buttons from -128 to +127.
*   ew
*     Vertical position of directional buttons from -128 to +127.
*   lx
*     Horizontal position of left stick from -128 to +127.
*   ly
*     Vertical position of left stick from -128 to +127.
*   rx
*     Horizontal position of right stick from -128 to +127.
*   ry
*     Vertical position of right stick from -128 to +127.
*
* RESULT
*
* NOTES
*
* SEE ALSO
*   HostGamepadList()
*   acepansion/plugin.h
*
*****************************************************************************
*
*/

ACEPANSION_API
VOID HostGamepadEvent(struct ACEpansionPlugin *plugin, UBYTE index, USHORT buttons, BYTE ns, BYTE ew, BYTE lx, BYTE ly, BYTE rx, BYTE ry)
{
    struct PluginData *myPlugin = (struct PluginData *)plugin;

    if(index == myPlugin->joy[0].idx)
    {
        myPlugin->joy[0].buttons = buttons;
        myPlugin->joy[0].ns = ns;
        myPlugin->joy[0].ew = ew;
        myPlugin->joy[0].lx = lx;
        myPlugin->joy[0].ly = ly;
        myPlugin->joy[0].rx = rx;
        myPlugin->joy[0].ry = ry;
    }
    if(index == myPlugin->joy[1].idx)
    {
        myPlugin->joy[1].buttons = buttons;
        myPlugin->joy[1].ns = ns;
        myPlugin->joy[1].ew = ew;
        myPlugin->joy[1].lx = lx;
        myPlugin->joy[1].ly = ly;
        myPlugin->joy[1].rx = rx;
        myPlugin->joy[1].ry = ry;
    }
}



/****** joysticks.acepansion/HostMouseEvent *********************************
*
* NAME
*   HostMouseEvent -- (V7) -- GUI

* SYNOPSIS
*   VOID HostMouseEvent(struct ACEpansionPlugin *plugin, UBYTE buttons, SHORT deltaX, SHORT deltaY)

* FUNCTION
*   Function that is called everytime a host's mouse event occured

* INPUTS
*   *plugin
*     Pointer to the plugin data structure to use.
*   buttons
*     State of the mouse buttons.
*     The mouse buttons bits can be decoded using the following flags:
*       ACE_HOSTMOUSEF_BUTTON_MAIN
*         State of main mouse button.
*       ACE_HOSTMOUSEF_BUTTON_SECOND
*         State of second mouse button.
*       ACE_HOSTMOUSEF_BUTTON_MIDDLE
*         State of middle mouse button.
*   deltaX
*     Horizontal mouse move until previous event.
*   deltaY
*     Vertical mouse move until previous event.
*
* RESULT
*
* NOTES
*
* SEE ALSO
      acepansion/plugin.h
*
*****************************************************************************
*
*/

ACEPANSION_API
VOID HostMouseEvent(UNUSED struct ACEpansionPlugin *plugin, UNUSED UBYTE buttons, UNUSED SHORT deltaX, UNUSED SHORT deltaY)
{
}



/****** joysticks.acepansion/HostLightDeviceDiodePulse **********************
*
* NAME
*   HostLightDeviceDiodePulse -- (V7) -- EMU
*
* SYNOPSIS
*   VOID HostLightDeviceDiodePulse(struct ACEpansionPlugin *plugin)
*
* FUNCTION
*   This function is called at each photo-diode pulse, when the beam from
*   the screen monitor is just in front of the light device photo-diode.
*
* INPUTS
*
* RESULT
*
* SEE ALSO
*   HostLightDeviceButton
*
*****************************************************************************
*
*/

ACEPANSION_API
VOID HostLightDeviceDiodePulse(UNUSED struct ACEpansionPlugin *plugin)
{
}



/****** joysticks.acepansion/HostLightDeviceButton **************************
*
* NAME
*   HostLightDeviceButton -- (V7) -- GUI
*
* SYNOPSIS
*   VOID HostLightDeviceButton(struct ACEpansionPlugin *plugin, BOOL pressed)
*
* FUNCTION
*   This function is called everytime the state of the light device button
*   is changing.
*
* INPUTS
*   pressed
*     New state of the light device button.
*
* RESULT
*
* NOTE
*   When a light device plugin is activated, ACE is mapping this button on
*   mouse main button, so that on-screen click will emulate light device
*   button.
*
* SEE ALSO
*   HostLightDeviceDiodePulse
*
*****************************************************************************
*
*/

ACEPANSION_API
VOID HostLightDeviceButton(UNUSED struct ACEpansionPlugin *plugin, UNUSED BOOL pressed)
{
}



/*
** Functions called from the GUI
*/

VOID Plugin_SetJoystick(struct ACEpansionPlugin *plugin, ULONG joystickId, ULONG selectedId)
{
    struct PluginData *myPlugin = (struct PluginData *)plugin;

    ObtainSemaphore(myPlugin->sema);

    myPlugin->joy[joystickId].idx = selectedId;
    FreeVec(myPlugin->joy[joystickId].preferredName);

    if(selectedId)
        myPlugin->joy[joystickId].preferredName = StrDup(myPlugin->list[selectedId]);
    else
        myPlugin->joy[joystickId].preferredName = NULL;

    ReleaseSemaphore(myPlugin->sema);
}

VOID Plugin_SetJoystickAnalogMode(struct ACEpansionPlugin *plugin, ULONG joystickId, BOOL analogMode)
{
    struct PluginData *myPlugin = (struct PluginData *)plugin;

    ObtainSemaphore(myPlugin->sema);
    myPlugin->joy[joystickId].analogMode = analogMode;
    ReleaseSemaphore(myPlugin->sema);
}

VOID Plugin_SetJoystickCombinedMode(struct ACEpansionPlugin *plugin, BOOL combinedMode)
{
    struct PluginData *myPlugin = (struct PluginData *)plugin;

    ObtainSemaphore(myPlugin->sema);
    myPlugin->combinedMode = combinedMode;
    ReleaseSemaphore(myPlugin->sema);
}
