/*
** Graphical User Interface using MUI
** for Joysticks ACEpansion
*/

#include <clib/alib_protos.h>

#include <proto/exec.h>
#include <proto/utility.h>
#include <proto/intuition.h>
#include <proto/muimaster.h>

#include <acepansion/lib_header.h>

#include "acepansion.h"
#include "interface.h"
#define CATCOMP_NUMBERS
#include "generated/locale_strings.h" /* catalog string ids */



struct Library *IntuitionBase;
struct Library *UtilityBase;
struct Library *MUIMasterBase;



struct MUI_CustomClass *MCC_JoysticksClass = NULL;



/*
** MCC_JoysticksClass definitions
*/
#define JoysticksObject NewObject(MCC_JoysticksClass->mcc_Class, NULL

#define MUIA_Joysticks_Plugin (TAG_USER | 0x0000) // [I..] struct ACEpansionPlugin *, Reference to the plugin managing the GUI (mandatory)

#define MUIM_Joysticks_Notify (TAG_USER | 0x0100) // struct MUIP_Joysticks_Notify, notify the plugin or the GUI about changes
    #define MUIV_Joysticks_Notify_ToGUI    0 // Notify the GUI
    #define MUIV_Joysticks_Notify_ToPlugin 1 // Notify the plugin
        // For MUIV_Joysticks_Notify_ToGUI only
        #define MUIV_Joysticks_Notify_List         10 // Notify about joysticks list and indexes
        // For MUIV_Joysticks_Notify_ToPlugin only
        #define MUIV_Joysticks_Notify_SelectedId0  11 // Notify about selected id for joystick 0
        #define MUIV_Joysticks_Notify_SelectedId1  12 // Notify about selected id for joystick 1
        // For both directions
        #define MUIV_Joysticks_Notify_CombinedMode 13 // Notify about combined mode
        #define MUIV_Joysticks_Notify_AnalogMode0  14 // Notify about joystick 0 analog mode
        #define MUIV_Joysticks_Notify_AnalogMode1  15 // Notify about joystick 1 analog mode

struct MUIP_Joysticks_Notify
{
    IPTR id;
    IPTR direction;
    IPTR type;
    union
    {
        // Type = MUIV_Joysticks_Notify_List
        struct
        {
            CONST_STRPTR *list;
            IPTR selectedId0;
            IPTR selectedId1;
        };
        // Type = MUIV_Joysticks_Notify_SelectedId0
        //        MUIV_Joysticks_Notify_SelectedId1
        IPTR selectedId;
        // Type = MUIV_Joysticks_Notify_AnalogMode0
        //        MUIV_Joysticks_Notify_AnalogMode1
        IPTR analogMode;
        // Type = MUIV_Joysticks_Notify_CombinedMode
        IPTR combinedMode;
    };
};

struct JoysticksData
{
    struct ACEpansionPlugin *plugin;

    Object *CYA_Select0;
    Object *CYA_Select1;
    Object *CHK_AnalogMode0;
    Object *CHK_AnalogMode1;
    Object *CHK_CombinedMode;
};



/*
** Helper functions
*/
static Object * makeJoystickObject(VOID)
{
    static CONST_STRPTR dummy[] = { "...", NULL };

    return CycleObject,
        MUIA_Cycle_Entries, dummy,
        MUIA_Cycle_Active, 0,
        MUIA_CycleChain, TRUE,
        End;
}

static Object * makeAnalogCheckmarkObject(Object **CHK_Analog)
{
    return HGroup,
        Child, *CHK_Analog = MUI_MakeObject(MUIO_Checkmark, GetString(MSG_CHK_ANALOG)),
        Child, HGroup,
            Child, LLabel(GetString(MSG_CHK_ANALOG)),
            Child, TextObject,
                MUIA_Font, MUIV_Font_Tiny,
                MUIA_Text_Contents, GetString(MSG_AMSTRAD_PLUS_ONLY),
                End,
            Child, HVSpace,
            End,
        MUIA_CycleChain, TRUE,
        End;
}

static VOID disableJoystick1(struct JoysticksData *data, BOOL disabled)
{
    // Disable joystick 1 gadgets in case of combined mode
    set(data->CYA_Select1, MUIA_Disabled, disabled);
    set(data->CHK_AnalogMode1, MUIA_Disabled, disabled);
}



/*
** MCC_JoysticksClass methods
*/
static IPTR mNotify(struct IClass *cl, Object *obj, struct MUIP_Joysticks_Notify *msg)
{
    struct JoysticksData *data = INST_DATA(cl,obj);

    switch(msg->direction)
    {
        case MUIV_Joysticks_Notify_ToGUI:
            switch(msg->type)
            {
                case MUIV_Joysticks_Notify_List:
                    // Update list
                    nnset(data->CYA_Select0, MUIA_Cycle_Entries, msg->list);
                    nnset(data->CYA_Select1, MUIA_Cycle_Entries, msg->list);
                    // Update indexes
                    nnset(data->CYA_Select0, MUIA_Cycle_Active, msg->selectedId0);
                    nnset(data->CYA_Select1, MUIA_Cycle_Active, msg->selectedId1);
                    break;
                case MUIV_Joysticks_Notify_CombinedMode:
                    nnset(data->CHK_CombinedMode, MUIA_Selected, msg->combinedMode);
                    disableJoystick1(data, msg->combinedMode);
                    break;
                case MUIV_Joysticks_Notify_AnalogMode0:
                    nnset(data->CHK_AnalogMode0, MUIA_Selected, msg->analogMode);
                case MUIV_Joysticks_Notify_AnalogMode1:
                    nnset(data->CHK_AnalogMode1, MUIA_Selected, msg->analogMode);
            }
            break;

        case MUIV_Joysticks_Notify_ToPlugin:
            switch(msg->type)
            {
                case MUIV_Joysticks_Notify_CombinedMode:
                    Plugin_SetJoystickCombinedMode(data->plugin, msg->combinedMode);
                    disableJoystick1(data, msg->combinedMode);
                    break;
                case MUIV_Joysticks_Notify_SelectedId0:
                    Plugin_SetJoystick(data->plugin, 0, msg->selectedId);
                    break;
                case MUIV_Joysticks_Notify_SelectedId1:
                    Plugin_SetJoystick(data->plugin, 1, msg->selectedId);
                    break;
                case MUIV_Joysticks_Notify_AnalogMode0:
                    Plugin_SetJoystickAnalogMode(data->plugin, 0, msg->analogMode);
                    break;
                case MUIV_Joysticks_Notify_AnalogMode1:
                    Plugin_SetJoystickAnalogMode(data->plugin, 1, msg->analogMode);
                    break;
            }
            break;
    }

    return 0;
}

// Workaround MUI issue not taking care of MUIA_Disabled updated before mSetup
static IPTR mSetup(struct IClass *cl,Object *obj, Msg msg)
{
    struct JoysticksData *data = INST_DATA(cl,obj);
    ULONG value;

    if(!(DoSuperMethodA(cl,obj,msg)))
        return FALSE;

    get(data->CHK_CombinedMode, MUIA_Selected, &value);
    set(data->CYA_Select1, MUIA_Disabled, value);
    set(data->CHK_AnalogMode1, MUIA_Disabled, value);

    return TRUE;
}

static IPTR mNew(struct IClass *cl, Object *obj, Msg msg)
{

    struct TagItem *tags,*tag;
    struct JoysticksData *data;

    struct ACEpansionPlugin *plugin = NULL;

    Object *CYA_Select0;
    Object *CYA_Select1;
    Object *CHK_AnalogMode0;
    Object *CHK_AnalogMode1;
    Object *CHK_CombinedMode;

    // Parse initial taglist
    for(tags=((struct opSet *)msg)->ops_AttrList; (tag=NextTagItem(&tags)); )
    {
        switch(tag->ti_Tag)
        {
            case MUIA_Joysticks_Plugin:
                plugin = (struct ACEpansionPlugin *)tag->ti_Data;
                break;                              
        }
    }

    if(!plugin) return 0;

    obj = (Object *)DoSuperNew(cl,obj,
        Child, VGroup,
            MUIA_Group_Columns, 2,
            MUIA_ShortHelp, GetString(MSG_GROUP_ASSIGN_HELP),
            GroupFrameT(GetString(MSG_GROUP_ASSIGN_TITLE)),
            Child, Label2(GetString(MSG_LABEL_JOY0)),
            Child, CYA_Select0 = makeJoystickObject(),

            Child, HVSpace,
            Child, makeAnalogCheckmarkObject(&CHK_AnalogMode0),

            Child, Label2(GetString(MSG_LABEL_JOY1)),
            Child, CYA_Select1 = makeJoystickObject(),

            Child, HVSpace,
            Child, makeAnalogCheckmarkObject(&CHK_AnalogMode1),
            End,
        Child, HGroup,
            MUIA_ShortHelp, GetString(MSG_GROUP_ADVANCED_HELP),
            GroupFrameT(GetString(MSG_GROUP_ADVANCED_TITLE)),
            Child, HGroup,
                Child, CHK_CombinedMode = MUI_MakeObject(MUIO_Checkmark, GetString(MSG_LABEL_CHK_COMBINED)),
                Child, LLabel(GetString(MSG_LABEL_CHK_COMBINED)),
                Child, HVSpace,
                MUIA_CycleChain, TRUE,
                End,
            End,
        TAG_MORE,((struct opSet *)msg)->ops_AttrList);

    if(!obj) return 0;

    data = INST_DATA(cl,obj);

    data->plugin = plugin;

    data->CYA_Select0 = CYA_Select0;
    data->CYA_Select1 = CYA_Select1;
    data->CHK_AnalogMode0 = CHK_AnalogMode0;
    data->CHK_AnalogMode1 = CHK_AnalogMode1;
    data->CHK_CombinedMode = CHK_CombinedMode;

    // Analog mode notify from GUI to plugin
    DoMethod(CHK_AnalogMode0, MUIM_Notify, MUIA_Selected, MUIV_EveryTime,
             obj, 4, MUIM_Joysticks_Notify, MUIV_Joysticks_Notify_ToPlugin,
                     MUIV_Joysticks_Notify_AnalogMode0, MUIV_TriggerValue);

    DoMethod(CHK_AnalogMode1, MUIM_Notify, MUIA_Selected, MUIV_EveryTime,
             obj, 4, MUIM_Joysticks_Notify, MUIV_Joysticks_Notify_ToPlugin,
                     MUIV_Joysticks_Notify_AnalogMode1, MUIV_TriggerValue);

    // Cycle select notify from GUI to plugin
    DoMethod(CYA_Select0, MUIM_Notify, MUIA_Cycle_Active, MUIV_EveryTime,
             obj, 4, MUIM_Joysticks_Notify, MUIV_Joysticks_Notify_ToPlugin,
                     MUIV_Joysticks_Notify_SelectedId0, MUIV_TriggerValue);

    DoMethod(CYA_Select1, MUIM_Notify, MUIA_Cycle_Active, MUIV_EveryTime,
             obj, 4, MUIM_Joysticks_Notify, MUIV_Joysticks_Notify_ToPlugin,
                     MUIV_Joysticks_Notify_SelectedId1, MUIV_TriggerValue);

    // Combined mode notify from GUI to plugin
    DoMethod(CHK_CombinedMode, MUIM_Notify, MUIA_Selected, MUIV_EveryTime,
             obj, 4, MUIM_Joysticks_Notify, MUIV_Joysticks_Notify_ToPlugin,
                     MUIV_Joysticks_Notify_CombinedMode, MUIV_TriggerValue);

    return (IPTR)obj;
}



/*
** MUI Custom Class dispatcher, creation & destruction
*/
DISPATCHER(JoysticksClass)
{
    switch (msg->MethodID)
    {
        case OM_NEW                : return(mNew   (cl,obj,(APTR)msg));
        case MUIM_Setup            : return(mSetup (cl,obj,(APTR)msg));
        case MUIM_Joysticks_Notify : return(mNotify(cl,obj,(APTR)msg));
    }
}
DISPATCHER_END



/*
** Function which is called to initialize commons just
** after the library was loaded into memory and prior
** to any other API call
**
** This is the good place to open our required libraries
** and create our MUI custom classes
*/
BOOL GUI_InitResources(VOID)
{
    UtilityBase = OpenLibrary("utility.library", 0L);
    IntuitionBase = OpenLibrary("intuition.library", 0L);
    MUIMasterBase = OpenLibrary("muimaster.library", 0L);

    if(UtilityBase && IntuitionBase && MUIMasterBase)
    {
        MCC_JoysticksClass = MUI_CreateCustomClass(
            NULL,
            MUIC_Group,
            NULL,
            sizeof(struct JoysticksData),
            DISPATCHER_REF(JoysticksClass));
    }

    return MCC_JoysticksClass != NULL;
}

/*
** Function which is called to free commons just before
** the library is expurged from memory
**
** This is the good place to close our required libraries
** and destroy our MUI custom classes
*/
VOID GUI_FreeResources(VOID)
{
    if(MCC_JoysticksClass)
        MUI_DeleteCustomClass(MCC_JoysticksClass);

    CloseLibrary(MUIMasterBase);
    CloseLibrary(IntuitionBase);
    CloseLibrary(UtilityBase);
}



/*
** Function which is called from CreatePlugin() to build the GUI
*/
Object * GUI_Create(struct ACEpansionPlugin *plugin)
{
    return JoysticksObject,
        MUIA_Joysticks_Plugin, plugin,
        End;
}

/*
** Function which is called from DeletePlugin() to delete the GUI
*/
VOID GUI_Delete(Object *gui)
{
    MUI_DisposeObject(gui);
}



/*
** Functions called from the plugin
*/
VOID GUI_SetJoystickList(Object *gui, STRPTR *list, ULONG selectedId0, ULONG selectedId1)
{
    DoMethod(gui, MUIM_Joysticks_Notify, MUIV_Joysticks_Notify_ToGUI,
             MUIV_Joysticks_Notify_List, list, selectedId0, selectedId1);
}

VOID GUI_SetJoystickCombinedMode(Object *gui, BOOL combinedMode)
{
    DoMethod(gui, MUIM_Joysticks_Notify, MUIV_Joysticks_Notify_ToGUI,
             MUIV_Joysticks_Notify_CombinedMode, combinedMode);
}

VOID GUI_SetJoystickAnalogMode(Object *gui, ULONG joystickId, BOOL analogMode)
{
    if(joystickId == 0)
    {
        DoMethod(gui, MUIM_Joysticks_Notify, MUIV_Joysticks_Notify_ToGUI,
                 MUIV_Joysticks_Notify_AnalogMode0, analogMode);
    }
    else if(joystickId == 1)
    {
        DoMethod(gui, MUIM_Joysticks_Notify, MUIV_Joysticks_Notify_ToGUI,
                 MUIV_Joysticks_Notify_AnalogMode1, analogMode);
    }
}

